from types import NoneType

import pytest
from core_10x.exec_control import CACHE_ONLY, CONVERT_VALUES_ON, DEBUG_ON
from core_10x.trait_definition import T
from core_10x.trait_method_error import TraitMethodError
from core_10x.traitable import Traitable
from core_10x.xnone import XNone, XNoneType


def _test_i(data_type: type, values, expected_values):
    assert len(values) == len(expected_values)

    class TestableTraitable(Traitable):
        xid: int = T(T.ID, default=1)
        value: data_type = T(1)  # noinspection PyTypeHints

    p = TestableTraitable(xid=1)
    for v, ev in zip(values, expected_values, strict=True):
        if isinstance(ev, type) and issubclass(ev, Exception):
            try:
                p.value = v
            except Exception as ex:
                if not isinstance(ex, ev):
                    pytest.fail(f'Got {ex} instead of {ev} value {v} for type {data_type}')
            else:
                pytest.fail(f'Expected exception {ev} for value {v} for type {data_type}; converted to {p.value}')
        else:
            p.value = v
            assert p.value == ev
            if ev is None:
                with pytest.raises(ValueError):
                    p.serialize_object()
            elif isinstance(ev, data_type):
                assert p.T.value.from_any(p.serialize_object()['value']) == ev


def _test(data_type, values, expected_types):
    with CACHE_ONLY():  # TODO: remove when db access is sorted
        _test_i(data_type, values, expected_types)


def generic_test(data_type, values, convert_expected, allowed_types=()):
    allowed_types = (data_type, NoneType, XNoneType, *allowed_types)
    debug_expected = [
        v if isinstance(v, allowed_types) else e if isinstance(e, Exception) else TypeError for v, e in zip(values, convert_expected, strict=True)
    ]
    _test(data_type, values, values)
    with DEBUG_ON():
        _test(data_type, values, debug_expected)
    with CONVERT_VALUES_ON():
        _test(data_type, values, convert_expected)
        with DEBUG_ON():
            _test(data_type, values, convert_expected)


def test_int_trait():
    data_type = int
    values = [180, 180.1, '180', 'abc', None, XNone, '']
    convert_expected = [180, 180, 180, TypeError, None, XNone, TypeError]
    generic_test(data_type, values, convert_expected)
    _test(data_type, values, values)


def test_float_trait():
    data_type = float
    values = [180, 180.1, '180', 'abc', None, XNone, '']
    convert_expected = [180.0, 180.1, 180.0, TypeError, None, XNone, TypeError]
    generic_test(data_type, values, convert_expected, (int, float))
    _test(data_type, values, values)


def test_bool_trait():
    data_type = bool
    values = [True, False, 1, 0, 'yes', 'no', 'abc', None, XNone, '', 'true', 'false', '1', '0', 'on', 'off']
    convert_expected = [
        True,
        False,
        True,
        False,
        True,
        False,
        TypeError,
        None,
        XNone,
        False,
        True,
        False,
        True,
        False,
        True,
        False,
    ]
    generic_test(data_type, values, convert_expected)


def test_datetime_trait():
    from datetime import datetime

    data_type = datetime
    values = [
        datetime(2020, 1, 1),
        '2020-01-01T01:02:03',
        '2020-01-01 00:00:00',
        '2020-01-01',
        1577836800,
        None,
        XNone,
        'abc',
        '',
        'Feb 1, 2020 1:2:3',
    ]
    convert_expected = [
        datetime(2020, 1, 1),
        datetime(2020, 1, 1, 1, 2, 3),
        datetime(2020, 1, 1),
        datetime(2020, 1, 1),
        TraitMethodError,
        None,
        XNone,
        TypeError,
        TypeError,
        datetime(2020, 2, 1, 1, 2, 3),
    ]
    generic_test(data_type, values, convert_expected)


def test_date_trait():
    from datetime import date

    data_type = date
    values = [
        date(2020, 1, 1),
        '2020-01-01T00:00:00',
        '2020-01-01 00:00:00',
        '2020-01-01',
        1577836800,
        None,
        XNone,
        'abc',
        '',
        'Feb 1, 2020',
    ]
    convert_expected = [
        date(2020, 1, 1),
        date(2020, 1, 1),
        date(2020, 1, 1),
        date(2020, 1, 1),
        TraitMethodError,
        None,
        XNone,
        TypeError,
        TypeError,
        date(2020, 2, 1),
    ]
    generic_test(data_type, values, convert_expected)



from __future__ import annotations

from core_10x.rc import RC
from core_10x.trait import Trait
from core_10x.traitable import Traitable
from matplotlib import colors

import rio


class StyleSheet(Traitable):
    sheet: dict
    text_style: rio.TextStyle

    def sheet_set(self, trait: Trait, value: dict) -> RC:
        # {'color': 'lightgreen', 'background-color': 'white', 'font-family': 'Helvetica', 'font-style': 'normal', 'font-weight': 'normal', 'border-width': '2px', 'border-style': '', 'border-color': 'blue'}
        style = dict(value)
        kw = {}
        if bg_color := style.pop('color', None):
            kw['fill'] = rio.Color.from_hex(colors.to_hex(bg_color))

        if font := style.pop('font-family', None):
            kw['font'] = rio.Font.from_google_fonts(font)

        if font_style := style.pop('font-style', None):
            if font_style != 'normal':
                kw[font_style] = True

        if font_weight := style.pop('font-weight', None):
            kw['font_weight'] = font_weight

        if kw:
            self.text_style = rio.TextStyle(**kw)

        return self.rc(style)

    def sheet_get(self):
        style = {}
        ts = self.text_style
        if ts:
            style['font-style'] = 'italic' if ts.italic else 'normal'
            style['font-weight'] = ts.font_weight or 'normal'
            if family := ts.font._google_fonts_name:  # TODO: better?
                style['font-family'] = family
            style['color'] = f'#{ts.fill.hex}'
        return style

    @staticmethod
    def rc(ss) -> RC:
        rc = RC(True)
        for style, value in ss.items():
            rc.add_error(f'Unsupported style: {style}: {value}')
        return rc



from __future__ import annotations

from typing import TYPE_CHECKING

from core_10x.attic.data_domain import GeneralDomain
from core_10x.py_class import PyClass
from core_10x.resource import TS_STORE

if TYPE_CHECKING:
    from core_10x.resource import ResourceRequirements


class PackageManifest:
    """
    To associate a subclass of Traitable to a particular Data Domain Category
    it must have a record in a package manifest file.
    Example:
        # module x.etrading_domain
        class ETradingDomain(DataDomain):
            GENERAL     = TS_STORE()
            ORDERS      = TS_STORE()
            ALGO_ORDERS = TS_STORE()
            ...

        #module x.y.order
        class Order(Traitable):
            ...

        class AlgoOrder(Order):
            ...


        #module x.y._package_manifest
        from x.etrading_domain import ETradingDomain as ET

        manifest = dict(        #-- variable manifest: dict must be defined as follows
            _category = ET.GENERAL,     #-- default category for Traitables in x.y package, if any

            order = dict(                   #-- specific association(s) for any module inside x.y package, if any
                _category = ET.ORDERS           #-- default category for Traitables in x.y.order, if any

                AlgoOrder = ET.ALGO_ORDERS      #-- specific association class x.y.order.AlgoOrder, if any
                ...
            ),
            ...
        )

        If no association could be found for a subclass of Traitable, the default association will be used: GeneralDomain.GENERAL
    """

    # fmt: off
    SYMBOL      = '_package_manifest.manifest'
    CATEGORY    = '_category'
    # fmt: on

    @staticmethod
    def _resource_requirements(cls) -> ResourceRequirements:
        module = cls.__module__
        parts = module.split('.')
        if len(parts) < 2:
            return None  # -- should be at least package.module, so it's most probably '__main__.Class'

        package = '.'.join(parts[:-1])
        man_def: dict = PyClass.find_symbol(f'{package}.{PackageManifest.SYMBOL}')
        if not man_def:
            return None

        short_module = parts[-1]
        module_entry = man_def.get(short_module)
        if module_entry:
            rr: ResourceRequirements = module_entry.get(cls.__name__)
            if not rr:
                rr = module_entry.get(PackageManifest.CATEGORY)
                if rr:
                    return rr

        return man_def.get(PackageManifest.CATEGORY)

    @staticmethod
    def resource_requirements(cls) -> ResourceRequirements:
        rr = PackageManifest._resource_requirements(cls) or GeneralDomain.GENERAL
        assert rr.resource_type is TS_STORE, f'{cls} - invalid category in the _package_manifest'
        return rr

    # -- TODO: handle parent classes, if ResourceRequirements for the class in not found



import pytest
from core_10x.exec_control import CONVERT_VALUES_OFF, CONVERT_VALUES_ON, DEBUG_OFF, DEBUG_ON, GRAPH_OFF, GRAPH_ON
from core_10x.traitable import Traitable


class Person(Traitable):
    """Test traitable with converter methods."""

    # Runtime traitable: omit T() so no storage context is required
    name: bytes
    status: str

    def name_from_str(self, trait, value: str) -> bytes:
        """Convert from string - title case the name."""
        return value.title().encode()

    def name_from_any_xstr(self, trait, value) -> bytes:
        """Convert from non-string - convert to string and title case."""
        return str(value).title().encode()

    def status_from_any_xstr(self, trait, value) -> bytes:
        """Convert any value to lowercase string."""
        return str(value).lower()


@pytest.mark.parametrize('graph_mode', [GRAPH_OFF, GRAPH_ON])
@pytest.mark.parametrize('debug_mode', [DEBUG_OFF, DEBUG_ON])
def test_converter_usage_with_execution_modes(graph_mode, debug_mode):
    """Test that converters work with different execution mode combinations."""
    with graph_mode(), debug_mode():
        with CONVERT_VALUES_ON():
            person = Person(name='john', status=True)
            assert person.name == b'John'
            assert person.status == 'true'
        with CONVERT_VALUES_OFF():
            person = Person()
            if debug_mode is DEBUG_ON:
                with pytest.raises(TypeError):
                    person.status = True
                with pytest.raises(TypeError):
                    person.name = 'john'
            else:
                person.name = 'john'
                person.status = True
                assert person.status is True
                assert person.name == 'john'
